/*

GSMSALG 0.1.1
by Luigi Auriemma
e-mail: aluigi@altervista.org
web:    http://aluigi.altervista.org


This is the algorithm used by Gamespy to contact the master servers
based on the Gamespy's procotol.

It supports the normal algorithm and enctype 2 (that is just a simple
operation more).

The following is an usage example for RogerWilco with a secure key
equal to "ABCDEF":

printf("%s\n", gsseckey("ABCDEF", "rW17Ko", 0));


And the following is an usage example for RogerWilco with a secure key
equal to "ABCDEF" using enctype 2:

printf("%s\n", gsseckey("ABCDEF", "rW17Ko", 2));


Simple 8-)

NOTE: the "normal" algorithm is used for a lot of things, in fact not
      only to retrieve servers'list but also to send heartbeats packets.
      I don't know other usage, if you know them contact me.
NOTE: the enctype 2 sends encoded data, so you must know the algorithm
      to decode them.

This software is covered by GNU/GPL
*/
#include <stdio.h>
#include <string.h>




unsigned char *gsseckey(unsigned char *secure, unsigned char *key, int enctype);
unsigned char gsvalfunc(unsigned char reg);







unsigned char *gsseckey(unsigned char *secure, unsigned char *key, int enctype) {
    static  unsigned char   validate[9];
    unsigned char           secbuf[6],
                            buff[256],
                            *ptr,
                            *ptrval,
                            *sec;
    unsigned long           i,
                            eax,
                            ebx,
                            ecx,
                            edx,
                            esi,
                            edi;


    ptr = buff;
    for(i = 0; i < 256; i++) {
        *ptr++ = i;
    }


    eax = 0;
    esi = 0;
    ptr = buff;
    for(i = 0; i < 256; i++) {
        if(eax == 6) eax = 0;

        ecx = key[eax];
        edx = *ptr;
        esi += ecx;
        edx += esi;
        edx &= 0x800000FF;
        if(edx & 0x80000000) {
            edx--;
            edx |= 0xffffff00;
        }
        eax++;
        esi = edx;
        ecx = *ptr;
        edx = buff[esi];
        *ptr = edx;
        buff[esi] = ecx;
        ptr++;
    }


    edi = 0;
    ebx = 0;
    memcpy(secbuf, secure, 6);
    sec = secbuf;

    for(i = 0; i < 6; i++) {
        ecx = *sec;
        edi += ecx + 1;
        edi &= 0x800000FF;
        if(edi & 0x80000000) {
            edi--;
            edi |= 0xffffff00;
        }
        ecx = buff[edi];
        ecx += ebx;
        if(ecx & 0x80000000) {
            ecx--;
            ecx |= 0xffffff00;
        }
        ebx = ecx;
        ebx &= 0xff;
        ecx = buff[edi];
        edx = buff[ebx];
        buff[edi] = edx;
        buff[ebx] = ecx;
        ecx += edx;
        ecx &= 0x800000FF;
        if(ecx & 0x80000000) {
            ecx--;
            ecx |= 0xffffff00;
        }
        edx = buff[ecx];
        ecx = *sec ^ edx;
        *sec = ecx;

        sec++;
    }


    if(enctype == 2) {
        ptr = key;
        sec = secbuf;
        for(i = 0; i < 6; i++) {
            *sec ^= *ptr;
            sec++;
            ptr++;
        }
    }


    sec = secbuf;
    ptrval = validate;
    for(i = 0; i < 2; i++) {
        eax = *sec++;
        ecx = *sec++;
        edx = eax;
        eax &= 3;
        ptrval[0] = gsvalfunc(edx >> 2);
        edx = ecx;
        eax <<= 4;
        eax += (edx >> 4);
        ecx &= 0xf;
        ptrval[1] = gsvalfunc(eax);
        eax = *sec++;
        edx = eax;
        eax &= 0x3f;
        ecx <<= 2;
        ptrval[3] = gsvalfunc(eax);
        ptrval[2] = gsvalfunc(ecx + (edx >> 6));
        ptrval += 4;
    }
    *ptrval = 0x00;
    return(validate);
}







unsigned char gsvalfunc(unsigned char reg) {
    if(reg < 0x1a) return(reg + 0x41);
    if(reg < 0x34) return(reg + 0x47);
    if(reg < 0x3e) return(reg - 4);
    if(reg == 0x3e) return(0x2b);
    if(reg == 0x3f) reg = 0;
        else reg = 1;
    reg--;
    reg &= 0x2f;
    return(reg);
}


